@@ -206,15 +206,10 @@ test_that("get/set option can roundtrip double options for statement", {
   )
 })
 
-test_that("void driver errors getting string option of incorrect type", {
+test_that("void driver errors getting string option of incompatible type", {
   db <- adbc_database_init(adbc_driver_void())
   adbc_database_set_options(db, list("some_key" = "some value"))
 
-  expect_error(
-    adbc_database_get_option_bytes(db, "some_key"),
-    class = "adbc_status_not_found"
-  )
-
   expect_error(
     adbc_database_get_option_int(db, "some_key"),
     class = "adbc_status_not_found"
@@ -226,6 +221,16 @@ test_that("void driver errors getting string option of incorrect type", {
   )
 })
 
+test_that("void driver can get string option of compatible type", {
+  db <- adbc_database_init(adbc_driver_void())
+  adbc_database_set_options(db, list("some_key" = "some value"))
+
+  expect_identical(
+    adbc_database_get_option_bytes(db, "some_key"),
+    charToRaw("some value")
+  )
+})
+
 test_that("void driver errors getting bytes option of incorrect type", {
   db <- adbc_database_init(adbc_driver_void())
   adbc_database_set_options(db, list("some_key" = charToRaw("some value")))
@@ -260,12 +265,20 @@ test_that("void driver errors getting integer option of incorrect type", {
     class = "adbc_status_not_found"
   )
 
-  expect_error(
+
+})
+
+test_that("void driver can get integer option of compatible type", {
+  db <- adbc_database_init(adbc_driver_void())
+  adbc_database_set_options(db, list("some_key" = 123L))
+
+  expect_identical(
     adbc_database_get_option_double(db, "some_key"),
-    class = "adbc_status_not_found"
+    123.0
   )
 })
 
+
 test_that("void driver errors getting double option of incorrect type", {
   db <- adbc_database_init(adbc_driver_void())
   adbc_database_set_options(db, list("some_key" = 123.4))
